/*
 *
 * Driver for the watch dog timer on the AMD Elan SC 520
 *
 * Elan SC530 Users manual, P 331 (section 19.1)
 * Register set manual, P 329 (Section 16.1)
 *
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/device.h>
#include <sys/types.h>
#include <sys/conf.h>
#include <sys/wdog.h>

#include <arch/i386/mmcr/wdog_mmcrvar.h>
#include <arch/i386/mmcr/mmcrreg.h>

#include <dev/sysmon/sysmonvar.h>

static int
wdog_mmcrmatch(struct device *parent, struct cfdata *match, void *aux);

static void
wdog_mmcrattach(struct device *parent, struct device *self, void *aux);

struct wdog_mmcr_softc {
	struct device 		sc_dev;
	/* the MMCR base address after we've mapped it in */
	caddr_t			mmcr_base;
	struct sysmon_wdog 	sc_smw;
	int			echobug_war;
};

struct cfattach wdog_mmcr_ca = {
	sizeof(struct wdog_mmcr_softc), wdog_mmcrmatch, wdog_mmcrattach
};

int	wdog_mmcr_setmode(struct sysmon_wdog *smw);
int	wdog_mmcr_tickle(struct sysmon_wdog *smw);

int 	wdog_ctl_write(struct wdog_mmcr_softc *sc, u_int16_t value);
int 	wdog_ctl_reset(struct wdog_mmcr_softc *sc);

extern struct cfdriver wdog_mmcr_cd;
extern caddr_t mmcr_base;

static int
wdog_mmcrmatch(struct device *parent, struct cfdata *match, void *aux)
{
	struct wdog_mmcr_attach_args *wdaa = aux;

	if (strcmp(wdaa->wdaa_busname, "wdog_mmcr") != 0)
		return(0);

	return(1);
}

/*
EXP_SEL bit field == 0x80 == 32.54 secs
*/

#define WDOG_ENABLE		0x8000
#define WDOG_RESET_ENABLE	0x4000
#define WDOG_30SEC_EXP_SEL	0x0080

static void
wdog_mmcrattach(struct device *parent, struct device *self, void *aux)
{
	struct wdog_mmcr_softc *sc = (void *)self;
	u_int16_t cpu_revid;
	u_int8_t reset_status;

	/* should already be set for us */
	sc->mmcr_base = mmcr_base;

	printf(": MMCR watchdog timer\n");

	sc->sc_smw.smw_name = "mmcr";
	sc->sc_smw.smw_cookie = sc;
	sc->sc_smw.smw_setmode = wdog_mmcr_setmode;
	sc->sc_smw.smw_tickle = wdog_mmcr_tickle;
	sc->sc_smw.smw_period = 32; /* actually 32.54  */
        if (sysmon_wdog_register(&sc->sc_smw) != 0)
		printf("%s: unable to register watchdog "
			"with sysmon\n", sc->sc_dev.dv_xname);

	/* check the cpu stepping to see if we need to work around a h/w bug */
	cpu_revid = *(u_int16_t *)(sc->mmcr_base + MMCR_CPU_REVID);
	
	printf("%s: product %d stepping %d.%d\n", sc->sc_dev.dv_xname,
		((cpu_revid & 0xff00) >> 8),
		((cpu_revid & 0x00f0) >> 4),
		(cpu_revid & 0x000f)
		);
	
	/* Processor Rev A1 has a bug that effects the watchdog timer.
	 * If the GP bus echo mode is enabled writing to the watchdog control
	 * register is blocked.
	 *
	 * the BIOS in the Net45[02]1 enabled GP bus echo mode for vairous
	 * reasons so we need to switch echo mode off when we talk to the
	 * watchdog.
	 *
	 */
	if (cpu_revid == 0x0001)
		sc->echobug_war = 1;
	
	/* check for reset status, and print if we reseted
	   due to watchdog reset...*/

#define MMCR_RESSTA_WDT_RST_DET_bit	8     /* watchdog timer reset*/
		
	if (reset_status & MMCR_RESSTA_WDT_RST_DET_bit) {
		printf("%s: WARNING: previous reset due to watchdog timer timeout\n",
			sc->sc_dev.dv_xname);

		*(u_int8_t *)(sc->mmcr_base + MMCR_RESSTA) =
			MMCR_RESSTA_WDT_RST_DET_bit;
	}
	
	/* setup the watchdog registers with some defaults */
	wdog_ctl_write(sc, (WDOG_RESET_ENABLE | WDOG_30SEC_EXP_SEL));
	wdog_ctl_reset(sc);
}

int
wdog_mmcr_setmode(struct sysmon_wdog *smw)
{
	struct wdog_mmcr_softc *sc = smw->smw_cookie;

	if ((smw->smw_mode & WDOG_MODE_MASK) == WDOG_MODE_DISARMED) {
		wdog_ctl_write(sc, (WDOG_RESET_ENABLE | WDOG_30SEC_EXP_SEL));
	} else {
		if (smw->smw_period == WDOG_PERIOD_DEFAULT)
			smw->smw_period = 32;
		else if (smw->smw_period != 32) {
			return (EINVAL);
		}
		wdog_ctl_write(sc, (WDOG_ENABLE | WDOG_RESET_ENABLE |
					WDOG_30SEC_EXP_SEL));
		wdog_mmcr_tickle(smw);
	}

	return(0);
}

int
wdog_mmcr_tickle(struct sysmon_wdog *smw)
{
	struct wdog_mmcr_softc *sc = smw->smw_cookie;

	wdog_ctl_reset(sc);

	return(0);
}

#define UNLOCK1 0x3333
#define UNLOCK2 0xCCCC

int wdog_ctl_write(struct wdog_mmcr_softc *sc, u_int16_t value)
{
	u_int8_t echo_mode;
	
	/* switch off GP bus echo mode if we need to. */
	if (sc->echobug_war) {
		echo_mode = *(u_int8_t *)(sc->mmcr_base + MMCR_GPIO_ECHOMODE);
		*(u_int8_t *)(sc->mmcr_base + MMCR_GPIO_ECHOMODE) =
			(echo_mode & 0xfe);
	}
	
	/* unlock the register */
	*(u_int16_t *)(sc->mmcr_base + MMCR_WDTMRCTL) = UNLOCK1;
	*(u_int16_t *)(sc->mmcr_base + MMCR_WDTMRCTL) = UNLOCK2;

	/* write whatever */
	*(u_int16_t *)(sc->mmcr_base + MMCR_WDTMRCTL) = value;

	/* switch GP bus echo mode back on if needed. */
	if (sc->echobug_war)
		*(u_int8_t *)(sc->mmcr_base + MMCR_GPIO_ECHOMODE) = echo_mode;

	return(0);
}

#define RESET1	0xAAAA
#define RESET2	0x5555

int wdog_ctl_reset(struct wdog_mmcr_softc *sc)
{
	u_int8_t echo_mode;
	
	/* switch off GP bus echo mode if we need to. */
	if (sc->echobug_war) {
		echo_mode = *(u_int8_t *)(sc->mmcr_base + MMCR_GPIO_ECHOMODE);
		*(u_int8_t *)(sc->mmcr_base + MMCR_GPIO_ECHOMODE) =
			(echo_mode & 0xfe);
	}

	/* reset the timer */
	*(u_int16_t *)(sc->mmcr_base + MMCR_WDTMRCTL) = RESET1;
	*(u_int16_t *)(sc->mmcr_base + MMCR_WDTMRCTL) = RESET2;

	/* switch GP bus echo mode back on if needed. */
	if (sc->echobug_war)
		*(u_int8_t *)(sc->mmcr_base + MMCR_GPIO_ECHOMODE) = echo_mode;
	
	return(0);
}


